"""
Functionality for reading Cosmo Skymed data into a SICD model.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Jarred Barber", "Wade Schwartzkopf")

import logging
from collections import OrderedDict
import os
import re
from typing import Tuple, Dict, BinaryIO, Union, Optional
from datetime import datetime

import numpy
from numpy.polynomial import polynomial
from scipy.constants import speed_of_light

import sarpy._extensions
from sarpy.compliance import bytes_to_string
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.blocks import Poly1DType, Poly2DType, RowColType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType, RadarModeType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType, \
    WaveformParametersType, ChanParametersType, TxStepType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType
from sarpy.io.complex.sicd_elements.GeoData import GeoDataType, SCPType
from sarpy.io.complex.sicd_elements.SCPCOA import SCPCOAType
from sarpy.io.complex.sicd_elements.Position import PositionType, XYZPolyType
from sarpy.io.complex.sicd_elements.Grid import GridType, DirParamType, WgtTypeType
from sarpy.io.complex.sicd_elements.Timeline import TimelineType, IPPSetType
from sarpy.io.complex.sicd_elements.ImageFormation import ImageFormationType, \
    RcvChanProcType
from sarpy.io.complex.sicd_elements.RMA import RMAType, INCAType
from sarpy.io.complex.sicd_elements.Radiometric import RadiometricType
from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.data_segment import HDF5DatasetSegment
from sarpy.io.general.format_function import ComplexFormatFunction
from sarpy.io.general.utils import get_seconds, parse_timestring, is_file_like, is_hdf5, h5py
from sarpy.io.complex.utils import fit_time_coa_polynomial, fit_position_xvalidation

logger = logging.getLogger(__name__)

_unhandled_id_text = 'Unhandled mission id `{}`'


##########
# helper functions

def _extract_attrs(h5_element, out=None):
    if out is None:
        out = OrderedDict()
    for the_key in h5_element.attrs:
        val = h5_element.attrs[the_key]
        out[the_key] = bytes_to_string(val) if isinstance(val, bytes) else val
    return out


def load_addin():
    """Check for a CSK addin module"""
    try:
        from sarpy.io.complex import csk_addin
        return csk_addin
    except ImportError:
        pass

    eps = sarpy._extensions.entry_points(group='sarpy.io.complex.csk')
    if not eps:
        return None
    if len(eps) > 1:
        raise RuntimeError("More than one CSK addin found")
    return list(eps)[0].load()


###########
# parser and interpreter for hdf5 attributes

class CSKDetails(object):
    """
    Parses and converts the Cosmo Skymed metadata
    """

    __slots__ = ('_file_name', '_mission_id', '_product_type')

    def __init__(self, file_name: str):
        """

        Parameters
        ----------
        file_name : str
        """

        if h5py is None:
            raise ImportError("Can't read Cosmo Skymed files, because the h5py dependency is missing.")

        if not os.path.isfile(file_name):
            raise SarpyIOError('Path {} is not a file'.format(file_name))

        with h5py.File(file_name, 'r') as hf:
            try:
                self._mission_id = hf.attrs['Mission ID'].decode('utf-8')
            except KeyError:
                raise SarpyIOError('The hdf file does not have the top level attribute "Mission ID"')
            try:
                self._product_type = hf.attrs['Product Type'].decode('utf-8')
            except KeyError:
                raise SarpyIOError('The hdf file does not have the top level attribute "Product Type"')

        if self._mission_id not in ['CSK', 'CSG', 'KMPS']:
            raise ValueError('Expected hdf5 attribute `Mission ID` should be one of "CSK", "CSG", or "KMPS"). '
                             'Got Mission ID = {}.'.format(self._mission_id))
        if 'SCS' not in self._product_type:
            raise ValueError('Expected hdf to contain complex products '
                             '(attribute `Product Type` which contains "SCS"). '
                             'Got Product Type = {}'.format(self._product_type))

        self._file_name = file_name

    @property
    def file_name(self) -> str:
        """
        str: the file name
        """

        return self._file_name

    @property
    def mission_id(self) -> str:
        """
        str: the mission id
        """

        return self._mission_id

    @property
    def product_type(self) -> str:
        """
        str: the product type
        """

        return self._product_type

    def _get_hdf_dicts(self) -> (dict, dict, Dict[str, Tuple[int, ...]], Dict[str, numpy.dtype], Dict[str, str]):
        with h5py.File(self._file_name, 'r') as hf:
            h5_dict = _extract_attrs(hf)
            band_dict = OrderedDict()
            shape_dict = OrderedDict()
            dtype_dict = OrderedDict()
            pixeltype_dict = OrderedDict()

            for gp_name in sorted(hf.keys()):
                if self._mission_id == 'CSG' and gp_name == 'LRHM':
                    continue  # this is extraneous

                gp = hf[gp_name]
                band_dict[gp_name] = OrderedDict()
                _extract_attrs(gp, out=band_dict[gp_name])

                if 'B0001' in gp:
                    beam_info = gp['B0001']
                    _extract_attrs(beam_info, out=band_dict[gp_name])

                if self._mission_id in ['CSK', 'KMPS']:
                    the_dataset = gp['SBI']
                elif self._mission_id == 'CSG':
                    the_dataset = gp['IMG']
                else:
                    raise ValueError(_unhandled_id_text.format(self._mission_id))
                _extract_attrs(the_dataset, out=band_dict[gp_name])

                shape_dict[gp_name] = the_dataset.shape[:2]
                dtype_dict[gp_name] = the_dataset.dtype
                if the_dataset.dtype.name == 'float32':
                    pixeltype_dict[gp_name] = 'RE32F_IM32F'
                elif the_dataset.dtype.name == 'int16':
                    pixeltype_dict[gp_name] = 'RE16I_IM16I'
                else:
                    raise ValueError(
                        'Got unexpected data type {}, name {}'.format(
                            the_dataset.dtype, the_dataset.dtype.name))

        return h5_dict, band_dict, shape_dict, dtype_dict, pixeltype_dict

    @staticmethod
    def _parse_pol(str_in: str) -> str:
        return '{}:{}'.format(str_in[0], str_in[1])

    def _get_polarization(self, h5_dict: dict, band_dict: dict, band_name: str) -> str:
        if 'Polarisation' in band_dict[band_name]:
            return band_dict[band_name]['Polarisation']
        elif 'Polarization' in h5_dict:
            return h5_dict['Polarization']
        else:
            raise ValueError(
                'Failed finding polarization for file {}\n\t'
                'mission id {} and band name {}'.format(self.file_name, self.mission_id, band_name))

    def _get_base_sicd(self, h5_dict: dict, band_dict: dict) -> SICDType:
        def get_collection_info() -> (dict, CollectionInfoType):
            acq_mode = h5_dict['Acquisition Mode'].upper()
            if self.mission_id == 'CSK':
                if acq_mode in ['HIMAGE', 'PINGPONG']:
                    mode_type = 'STRIPMAP'
                elif acq_mode in ['WIDEREGION', 'HUGEREGION']:
                    # scansar, processed as stripmap
                    mode_type = 'STRIPMAP'
                elif acq_mode in ['ENHANCED SPOTLIGHT', 'SMART']:
                    mode_type = 'DYNAMIC STRIPMAP'
                else:
                    logger.warning('Got unexpected acquisition mode {}'.format(acq_mode))
                    mode_type = 'DYNAMIC STRIPMAP'
            elif self.mission_id == 'KMPS':
                if acq_mode in ['STANDARD', 'ENHANCED STANDARD']:
                    mode_type = 'STRIPMAP'
                elif acq_mode in ['WIDE SWATH', 'ENHANCED WIDE SWATH']:
                    # scansar, processed as stripmap
                    mode_type = 'STRIPMAP'
                elif acq_mode in ['HIGH RESOLUTION', 'ENHANCED HIGH RESOLUTION', 'ULTRA HIGH RESOLUTION']:
                    # "spotlight"
                    mode_type = 'DYNAMIC STRIPMAP'
                else:
                    logger.warning('Got unexpected acquisition mode {}'.format(acq_mode))
                    mode_type = 'DYNAMIC STRIPMAP'
            elif self.mission_id == 'CSG':
                if acq_mode.startswith('SPOTLIGHT'):
                    mode_type = 'DYNAMIC STRIPMAP'
                elif acq_mode in ['STRIPMAP', 'QUADPOL']:
                    mode_type = "STRIPMAP"
                else:
                    logger.warning(
                        'Got unhandled acquisition mode {},\n\t'
                        'setting to DYNAMIC STRIPMAP'.format(acq_mode))
                    mode_type = 'DYNAMIC STRIPMAP'
            else:
                raise ValueError(_unhandled_id_text.format(self._mission_id))

            start_time_dt = collect_start.astype('datetime64[s]').astype(datetime)
            date_str = start_time_dt.strftime('%d%b%y').upper()
            time_str = start_time_dt.strftime('%H%M%S') + 'Z'
            core_name = '{}_{}_{}'.format(date_str, h5_dict['Satellite ID'], time_str)
            collect_info = CollectionInfoType(
                Classification='UNCLASSIFIED',
                CollectorName=h5_dict['Satellite ID'],
                CoreName=core_name,
                CollectType='MONOSTATIC',
                RadarMode=RadarModeType(ModeID=h5_dict['Multi-Beam ID'],
                                        ModeType=mode_type))
            return collect_info

        def get_image_creation() -> ImageCreationType:
            from sarpy.__about__ import __version__
            return ImageCreationType(
                DateTime=parse_timestring(h5_dict['Product Generation UTC'], precision='ns'),
                Site=h5_dict['Processing Centre'],
                Application='L0: `{}`, L1: `{}`'.format(
                    h5_dict.get('L0 Software Version', 'NONE'),
                    h5_dict.get('L1A Software Version', 'NONE')),
                Profile='sarpy {}'.format(__version__))

        def get_grid() -> GridType:
            def get_wgt_type(weight_name, coefficient, direction):
                if weight_name == 'GENERAL_COSINE':
                    # probably only for kompsat?
                    weight_name = 'HAMMING'
                    coefficient = 1-coefficient
                if coefficient is None:
                    params = None
                else:
                    params = {'COEFFICIENT': '{0:0.17E}'.format(coefficient)}
                out = WgtTypeType(WindowName=weight_name, Parameters=params)
                if weight_name != 'HAMMING':
                    logger.warning(
                        'Got unexpected weight scheme {} for {}.\n\t'
                        'The weighting will not be properly populated.'.format(weight_name, direction))
                return out

            if re.sub(' ', '', h5_dict['Projection ID']).upper() == 'SLANTRANGE/AZIMUTH':
                image_plane = 'SLANT'
                gr_type = 'RGZERO'
            else:
                image_plane = 'GROUND'
                gr_type = 'PLANE'
            # Row
            row_window_name = h5_dict['Range Focusing Weighting Function'].rstrip().upper()
            row_coefficient = h5_dict.get('Range Focusing Weighting Coefficient', None)
            row_weight = get_wgt_type(row_window_name, row_coefficient, 'Row')
            row = DirParamType(Sgn=-1,
                               KCtr=2*center_frequency/speed_of_light,
                               DeltaKCOAPoly=Poly2DType(Coefs=[[0, ], ]),
                               WgtType=row_weight)
            # Col
            col_window_name = h5_dict['Azimuth Focusing Weighting Function'].rstrip().upper()
            col_coefficient = h5_dict.get('Azimuth Focusing Weighting Coefficient', None)
            col_weight = get_wgt_type(col_window_name, col_coefficient, 'Col')
            col = DirParamType(Sgn=-1, KCtr=0, WgtType=col_weight)
            return GridType(ImagePlane=image_plane, Type=gr_type, Row=row, Col=col)

        def get_timeline() -> TimelineType:
            # NB: IPPEnd must be set, but will be replaced
            return TimelineType(CollectStart=collect_start,
                                CollectDuration=duration,
                                IPP=[IPPSetType(index=0, TStart=0, TEnd=0, IPPStart=0, IPPEnd=0), ])

        def get_position() -> PositionType:
            T = h5_dict['State Vectors Times']  # in seconds relative to ref time
            T += ref_time_offset
            Pos = h5_dict['ECEF Satellite Position']
            Vel = h5_dict['ECEF Satellite Velocity']
            P_x, P_y, P_z = fit_position_xvalidation(T, Pos, Vel, max_degree=8)
            return PositionType(ARPPoly=XYZPolyType(X=P_x, Y=P_y, Z=P_z))

        def get_radar_collection() -> RadarCollectionType:
            tx_pols = []
            chan_params = []

            if self.mission_id == 'CSG' and len(band_dict) == 1 and \
                    h5_dict['Acquisition Mode'].upper() == 'QUADPOL':
                # it seems like 2nd generation files only contain one polarization
                pols = ['HH', 'HV', 'VH', 'VV']
                tx_pols.extend([pol[0] for pol in pols])
                chan_params.extend([
                    ChanParametersType(TxRcvPolarization=self._parse_pol(pol), index=i+1)
                    for i, pol in enumerate(pols)])
            else:
                for i, bdname in enumerate(band_dict):
                    pol = self._get_polarization(h5_dict, band_dict, bdname)
                    tx_pols.append(pol[0])
                    chan_params.append(ChanParametersType(TxRcvPolarization=self._parse_pol(pol), index=i+1))

            if len(tx_pols) == 1:
                return RadarCollectionType(RcvChannels=chan_params, TxPolarization=tx_pols[0])
            else:
                return RadarCollectionType(RcvChannels=chan_params,
                                           TxPolarization='SEQUENCE',
                                           TxSequence=[TxStepType(TxPolarization=pol,
                                                                  index=i+1) for i, pol in enumerate(tx_pols)])

        def get_image_formation() -> ImageFormationType:
            return ImageFormationType(ImageFormAlgo='RMA',
                                      TStartProc=0,
                                      TEndProc=duration,
                                      STBeamComp='NO',
                                      ImageBeamComp='SV',
                                      AzAutofocus='NO',
                                      RgAutofocus='NO',
                                      RcvChanProc=RcvChanProcType(NumChanProc=1,
                                                                  PRFScaleFactor=1))

        def get_rma() -> RMAType:
            inca = INCAType(FreqZero=center_frequency)
            return RMAType(RMAlgoType='OMEGA_K',
                           INCA=inca)

        def get_scpcoa() -> SCPCOAType:
            return SCPCOAType(SideOfTrack=h5_dict['Look Side'][0:1].upper())

        # some common use parameters
        center_frequency = h5_dict['Radar Frequency']
        # relative times in csk are wrt some reference time - for sicd they should be relative to start time
        collect_start = parse_timestring(h5_dict['Scene Sensing Start UTC'], precision='ns')
        collect_end = parse_timestring(h5_dict['Scene Sensing Stop UTC'], precision='ns')
        duration = get_seconds(collect_end, collect_start, precision='ns')
        ref_time = parse_timestring(h5_dict['Reference UTC'], precision='ns')
        ref_time_offset = get_seconds(ref_time, collect_start, precision='ns')

        # assemble our pieces
        collection_info = get_collection_info()
        image_creation = get_image_creation()
        grid = get_grid()
        timeline = get_timeline()
        position = get_position()
        radar_collection = get_radar_collection()
        image_formation = get_image_formation()
        rma = get_rma()
        scpcoa = get_scpcoa()
        sicd = SICDType(
            CollectionInfo=collection_info,
            ImageCreation=image_creation,
            Grid=grid,
            Timeline=timeline,
            Position=position,
            RadarCollection=radar_collection,
            ImageFormation=image_formation,
            RMA=rma,
            SCPCOA=scpcoa)
        return sicd

    def _get_dop_poly_details(self,
                              h5_dict: dict,
                              band_dict: dict,
                              band_name: str) -> (float, float, numpy.ndarray, numpy.ndarray, numpy.ndarray):
        def strip_poly(arr: numpy.ndarray) -> numpy.ndarray:
            # strip worthless (all zero) highest order terms
            # find last non-zero index
            last_ind = arr.size
            for i in range(arr.size-1, -1, -1):
                if arr[i] != 0:
                    break
                last_ind = i
            if last_ind == 0:
                return numpy.array([0, ], dtype=arr.dtype)
            return arr[:last_ind]

        dop_rate_poly_rg = h5_dict.get('Doppler Rate vs Range Time Polynomial', None)
        if dop_rate_poly_rg is None:
            dop_rate_poly_rg = band_dict[band_name].get('Doppler Rate vs Range Time Polynomial', None)
        if dop_rate_poly_rg is None:
            raise ValueError('No Doppler Rate Range Time polynomial found')
        dop_rate_poly_rg = strip_poly(dop_rate_poly_rg)

        if self._mission_id in ['CSK', 'KMPS']:
            az_ref_time = h5_dict['Azimuth Polynomial Reference Time']  # seconds
            rg_ref_time = h5_dict['Range Polynomial Reference Time']
            dop_poly_az = strip_poly(h5_dict['Centroid vs Azimuth Time Polynomial'])
            dop_poly_rg = strip_poly(h5_dict['Centroid vs Range Time Polynomial'])
        elif self._mission_id == 'CSG':
            az_ref_time_nozd = band_dict[band_name]['Azimuth Polynomial Reference Time']
            first_time = band_dict[band_name]['Azimuth First Time']
            last_time = band_dict[band_name]['Azimuth Last Time']
            az_fit_times = numpy.linspace(first_time, last_time, num=11)

            geom_dop_cent_poly = band_dict[band_name]['Doppler Centroid vs Azimuth Time Polynomial - RAW']

            dop_rate_poly = h5_dict.get('Doppler Rate vs Azimuth Time Polynomial', None)
            if dop_rate_poly is None:
                dop_rate_poly = band_dict[band_name].get('Doppler Rate vs Azimuth Time Polynomial', None)
            if dop_rate_poly is None:
                raise ValueError('No Doppler Rate Range Time polynomial found')

            centroid_values = polynomial.polyval(az_fit_times - az_ref_time_nozd, geom_dop_cent_poly)
            rate_values = polynomial.polyval(az_fit_times - az_ref_time_nozd, dop_rate_poly)
            zd_times = az_fit_times - centroid_values / rate_values
            az_ref_time = band_dict[band_name]['Azimuth Polynomial Reference Time - ZD']
            dop_poly_az = strip_poly(polynomial.polyfit(zd_times - az_ref_time, centroid_values, 4))

            rg_ref_time = band_dict[band_name]['Range Polynomial Reference Time']
            dop_poly_rg = strip_poly(band_dict[band_name]['Doppler Centroid vs Range Time Polynomial'])
        else:
            raise ValueError(_unhandled_id_text.format(self._mission_id))
        return az_ref_time, rg_ref_time, dop_poly_az, dop_poly_rg, dop_rate_poly_rg

    def _get_band_specific_sicds(self,
                                 base_sicd: SICDType,
                                 h5_dict: dict,
                                 band_dict: dict,
                                 shape_dict: dict,
                                 pixeltype_dict: dict) -> Dict[str, SICDType]:
        def update_scp_prelim(sicd: SICDType, band_name: str) -> None:
            if self._mission_id in ['CSK', 'KMPS']:
                LLH = band_dict[band_name]['Centre Geodetic Coordinates']
            elif self._mission_id == 'CSG':
                LLH = h5_dict['Scene Centre Geodetic Coordinates']
            else:
                raise ValueError(_unhandled_id_text.format(self._mission_id))
            sicd.GeoData = GeoDataType(SCP=SCPType(LLH=LLH))  # EarthModel & ECF will be populated

        def update_image_data(sicd: SICDType, band_name: str) -> (float, float, float, float, int):
            cols, rows = shape_dict[band_name]
            # zero doppler time of first/last columns
            t_az_first_time = band_dict[band_name]['Zero Doppler Azimuth First Time']
            t_az_last_time = band_dict[band_name]['Zero Doppler Azimuth Last Time']
            t_ss_az_s = band_dict[band_name]['Line Time Interval']
            t_use_sign2 = 1
            if h5_dict['Look Side'].upper() == 'LEFT':
                t_use_sign2 = -1
                t_az_first_time, t_az_last_time = t_az_last_time, t_az_first_time
            # zero doppler time of first row
            t_rg_first_time = band_dict[band_name]['Zero Doppler Range First Time']
            # row spacing in range time (seconds)
            t_ss_rg_s = band_dict[band_name]['Column Time Interval']

            sicd.ImageData = ImageDataType(NumRows=rows,
                                           NumCols=cols,
                                           FirstRow=0,
                                           FirstCol=0,
                                           FullImage=(rows, cols),
                                           PixelType=pixeltype_dict[band_name],
                                           SCPPixel=RowColType(Row=int(rows/2),
                                                               Col=int(cols/2)))
            return t_rg_first_time, t_ss_rg_s, t_az_first_time, t_ss_az_s, t_use_sign2

        def check_switch_state() -> (int, Poly1DType):
            use_sign = 1 if t_dop_rate_poly_rg[0] < 0 else -1
            return use_sign, Poly1DType(Coefs=use_sign*t_dop_rate_poly_rg)

        def update_timeline(sicd: SICDType, band_name: str) -> None:
            prf = band_dict[band_name]['PRF']
            duration = sicd.Timeline.CollectDuration
            ipp_el = sicd.Timeline.IPP[0]
            ipp_el.IPPEnd = round(duration*prf) - 1
            ipp_el.TEnd = duration
            ipp_el.IPPPoly = Poly1DType(Coefs=(0, prf))

        def update_radar_collection(sicd: SICDType, band_name: str) -> None:
            ind = None
            for the_chan_index, chan in enumerate(sicd.RadarCollection.RcvChannels):
                if chan.TxRcvPolarization == polarization:
                    ind = the_chan_index
                    break
            if ind is None:
                raise ValueError('Failed to find receive channel for polarization {}'.format(polarization))

            chirp_length = band_dict[band_name]['Range Chirp Length']
            chirp_rate = abs(band_dict[band_name]['Range Chirp Rate'])
            sample_rate = band_dict[band_name]['Sampling Rate']
            ref_dechirp_time = band_dict[band_name].get('Reference Dechirping Time', 0)  # TODO: is this right?
            win_length = band_dict[band_name]['Echo Sampling Window Length']
            rcv_fm_rate = 0 if numpy.isnan(ref_dechirp_time) else chirp_rate
            band_width = chirp_length*chirp_rate
            fr_min = center_frequency - 0.5*band_width
            fr_max = center_frequency + 0.5*band_width
            sicd.RadarCollection.TxFrequency = (fr_min, fr_max)
            sicd.RadarCollection.Waveform = [
                WaveformParametersType(index=0,
                                       TxPulseLength=chirp_length,
                                       TxRFBandwidth=band_width,
                                       TxFreqStart=fr_min,
                                       TxFMRate=chirp_rate,
                                       ADCSampleRate=sample_rate,
                                       RcvFMRate=rcv_fm_rate,
                                       RcvWindowLength=win_length/sample_rate), ]
            sicd.ImageFormation.RcvChanProc.ChanIndices = [ind+1, ]
            sicd.ImageFormation.TxFrequencyProc = (fr_min, fr_max)

        def update_rma_and_grid(sicd: SICDType, band_name: str) -> None:
            rg_scp_time = rg_first_time + (ss_rg_s*sicd.ImageData.SCPPixel.Row)
            az_scp_time = az_first_time + (use_sign2*ss_az_s*sicd.ImageData.SCPPixel.Col)
            r_ca_scp = rg_scp_time*speed_of_light/2
            sicd.RMA.INCA.R_CA_SCP = r_ca_scp
            # compute DRateSFPoly
            scp_ca_time = az_scp_time + ref_time_offset
            vel_poly = sicd.Position.ARPPoly.derivative(der_order=1, return_poly=True)
            vel_ca_vec = vel_poly(scp_ca_time)
            vel_ca_sq = numpy.sum(vel_ca_vec*vel_ca_vec)
            vel_ca = numpy.sqrt(vel_ca_sq)
            r_ca = numpy.array([r_ca_scp, 1.], dtype=numpy.float64)
            dop_rate_poly_rg_shifted = dop_rate_poly_rg.shift(
                rg_ref_time-rg_scp_time, alpha=ss_rg_s/row_ss, return_poly=False)
            drate_sf_poly = -(polynomial.polymul(dop_rate_poly_rg_shifted, r_ca) *
                              speed_of_light/(2*center_frequency*vel_ca_sq))
            # update grid.row
            sicd.Grid.Row.SS = row_ss
            sicd.Grid.Row.ImpRespBW = row_bw
            sicd.Grid.Row.DeltaK1 = -0.5 * row_bw
            sicd.Grid.Row.DeltaK2 = 0.5 * row_bw
            # update grid.col
            col_ss = abs(vel_ca*ss_az_s*drate_sf_poly[0])
            sicd.Grid.Col.SS = col_ss
            if self.mission_id == 'CSK':
                col_bw = min(band_dict[band_name]['Azimuth Focusing Transition Bandwidth']*ss_az_s, 1) / col_ss
            elif self.mission_id in ['CSG', 'KMPS']:
                col_bw = min(band_dict[band_name]['Azimuth Focusing Bandwidth']*ss_az_s, 1) / col_ss
            else:
                raise ValueError('Got unhandled mission_id {}'.format(self.mission_id))
            sicd.Grid.Col.ImpRespBW = col_bw
            # update inca
            sicd.RMA.INCA.DRateSFPoly = Poly2DType(Coefs=numpy.reshape(drate_sf_poly, (-1, 1)))
            sicd.RMA.INCA.TimeCAPoly = Poly1DType(Coefs=[scp_ca_time, use_sign2*ss_az_s/col_ss])
            # compute DopCentroidPoly & DeltaKCOAPoly
            dop_centroid_poly = numpy.zeros((dop_poly_rg.order1+1, dop_poly_az.order1+1), dtype=numpy.float64)
            dop_centroid_poly[0, 0] = dop_poly_rg(rg_scp_time-rg_ref_time) + \
                dop_poly_az(az_scp_time-az_ref_time) - \
                0.5*(dop_poly_rg[0] + dop_poly_az[0])
            dop_poly_rg_shifted = dop_poly_rg.shift(rg_ref_time-rg_scp_time, alpha=ss_rg_s/row_ss)
            dop_poly_az_shifted = dop_poly_az.shift(az_ref_time-az_scp_time, alpha=ss_az_s/col_ss)
            dop_centroid_poly[1:, 0] = dop_poly_rg_shifted[1:]
            dop_centroid_poly[0, 1:] = dop_poly_az_shifted[1:]
            sicd.RMA.INCA.DopCentroidPoly = Poly2DType(Coefs=dop_centroid_poly)
            sicd.RMA.INCA.DopCentroidCOA = True
            sicd.Grid.Col.DeltaKCOAPoly = Poly2DType(Coefs=use_sign*dop_centroid_poly*ss_az_s/col_ss)
            # fit TimeCOAPoly
            sicd.Grid.TimeCOAPoly = fit_time_coa_polynomial(
                sicd.RMA.INCA, sicd.ImageData, sicd.Grid, dop_rate_poly_rg_shifted, poly_order=2)

            csk_addin = load_addin()
            if csk_addin is not None:
                csk_addin.check_sicd(sicd, self.mission_id, h5_dict)

        def update_radiometric(sicd: SICDType, band_name: str) -> None:
            if self.mission_id in ['KMPS', 'CSG']:
                # TODO: skipping for now - strange results for flag == 77. Awaiting gidance - see Wade.
                return
            if h5_dict['Range Spreading Loss Compensation Geometry'] != 'NONE':
                slant_range = h5_dict['Reference Slant Range']
                exp = h5_dict['Reference Slant Range Exponent']
                sf = slant_range**(2*exp)
                rsf = h5_dict['Rescaling Factor']
                sf /= rsf * rsf
                if h5_dict.get('Calibration Constant Compensation Flag', None) == 0:
                    cal = band_dict[band_name]['Calibration Constant']
                    sf /= cal
                sicd.Radiometric = RadiometricType(BetaZeroSFPoly=Poly2DType(Coefs=[[sf, ], ]))

        def update_geodata(sicd: SICDType) -> None:
            scp_pixel = [sicd.ImageData.SCPPixel.Row, sicd.ImageData.SCPPixel.Col]
            ecf = sicd.project_image_to_ground(scp_pixel, projection_type='HAE')
            sicd.update_scp(ecf, coord_system='ECF')

            SCP = sicd.GeoData.SCP.ECF.get_array(dtype='float64')
            scp_time = sicd.RMA.INCA.TimeCAPoly[0]
            ca_pos = sicd.Position.ARPPoly(scp_time)
            RG = SCP - ca_pos
            sicd.RMA.INCA.R_CA_SCP = numpy.linalg.norm(RG)

        out = {}
        center_frequency = h5_dict['Radar Frequency']
        # relative times in csk are wrt some reference time - for sicd they should be relative to start time
        collect_start = parse_timestring(h5_dict['Scene Sensing Start UTC'], precision='ns')
        ref_time = parse_timestring(h5_dict['Reference UTC'], precision='ns')
        ref_time_offset = get_seconds(ref_time, collect_start, precision='ns')

        for bd_name in band_dict:
            polarization = self._parse_pol(self._get_polarization(h5_dict, band_dict, bd_name))
            az_ref_time, rg_ref_time, t_dop_poly_az, t_dop_poly_rg, t_dop_rate_poly_rg = \
                self._get_dop_poly_details(h5_dict, band_dict, bd_name)
            dop_poly_az = Poly1DType(Coefs=t_dop_poly_az)
            dop_poly_rg = Poly1DType(Coefs=t_dop_poly_rg)

            t_sicd = base_sicd.copy()
            t_sicd.ImageFormation.TxRcvPolarizationProc = polarization

            row_bw = band_dict[bd_name]['Range Focusing Bandwidth']*2/speed_of_light
            row_ss = band_dict[bd_name]['Column Spacing']
            rg_first_time, ss_rg_s, az_first_time, ss_az_s, use_sign2 = update_image_data(t_sicd, bd_name)
            use_sign, dop_rate_poly_rg = check_switch_state()
            update_timeline(t_sicd, bd_name)
            update_radar_collection(t_sicd, bd_name)
            update_rma_and_grid(t_sicd, bd_name)
            update_radiometric(t_sicd, bd_name)

            update_scp_prelim(t_sicd, bd_name)  # set preliminary value for SCP (required for projection)
            update_geodata(t_sicd)
            t_sicd.derive()
            # t_sicd.populate_rniirs(override=False)
            out[bd_name] = t_sicd
        return out

    @staticmethod
    def _get_symmetry(h5_dict: dict) -> (Optional[Tuple[int, ...]], Tuple[int, ...]):
        reverse_axes = []

        line_order = h5_dict['Lines Order'].upper()
        look_side = h5_dict['Look Side'].upper()
        symm_0 = ((line_order == 'EARLY-LATE') != (look_side == 'RIGHT'))
        if symm_0:
            reverse_axes.append(0)

        column_order = h5_dict['Columns Order'].upper()
        symm_1 = column_order != 'NEAR-FAR'
        if symm_1:
            reverse_axes.append(1)

        transpose_axes = (1, 0, 2)
        return tuple(reverse_axes), transpose_axes

    def get_sicd_collection(self) -> (
            Dict[str, SICDType], Dict[str, Tuple[int, ...]], Optional[Tuple[int, ...]], Tuple[int, ...]):
        """
        Get the sicd collection for the bands.

        Returns
        -------
        sicd_dict : Dict[str, SICDType]
            Of the form {band_name: sicd}
        shape_dict : Dict[str, Tuple[int, ...]]
            Of the form {band_name: shape}
        dtype_dict : Dict[str, numpy.dtype]
            Of the form {band_name: data type string}
        reverse_axes : Optional[Tuple[int, ...]]
        transpose_axes : Tuple[int, ...]
        """

        h5_dict, band_dict, shape_dict, dtype_dict, pixeltype_dict = self._get_hdf_dicts()
        base_sicd = self._get_base_sicd(h5_dict, band_dict)
        # noinspection PyTypeChecker
        return (self._get_band_specific_sicds(base_sicd, h5_dict, band_dict, shape_dict, pixeltype_dict), shape_dict, dtype_dict) + self._get_symmetry(h5_dict)


################
# The CSK reader


class CSKReader(SICDTypeReader):
    """
    A Cosmo SkyMed 1st or 2nd generation SLC reader implementation.

    **Changed in version 1.3.0** for reading changes.
    """

    __slots__ = ('_csk_details', )

    def __init__(self, csk_details):
        """

        Parameters
        ----------
        csk_details : str|CSKDetails
            file name or CSKDetails object
        """

        if isinstance(csk_details, str):
            csk_details = CSKDetails(csk_details)
        if not isinstance(csk_details, CSKDetails):
            raise TypeError('The input argument for a CSKReader must be a '
                            'filename or CSKDetails object')
        self._csk_details = csk_details
        sicd_data, shape_dict, dtype_dict, reverse_axes, transpose_axes = csk_details.get_sicd_collection()
        data_segments = []
        sicds = []
        for band_name in sicd_data:
            if self._csk_details.mission_id in ['CSK', 'KMPS']:
                the_band = '{}/SBI'.format(band_name)
            elif self._csk_details.mission_id == 'CSG':
                the_band = '{}/IMG'.format(band_name)
            else:
                raise ValueError(_unhandled_id_text.format(self._csk_details.mission_id))

            sicds.append(sicd_data[band_name])
            basic_shape = shape_dict[band_name]
            data_segments.append(
                HDF5DatasetSegment(
                    csk_details.file_name, the_band,
                    formatted_dtype='complex64', formatted_shape=(basic_shape[1], basic_shape[0]),
                    reverse_axes=reverse_axes, transpose_axes=transpose_axes,
                    format_function=ComplexFormatFunction(raw_dtype=dtype_dict[band_name], order='IQ', band_dimension=2),
                    close_file=True))

        SICDTypeReader.__init__(self, data_segments, sicds, close_segments=True)
        self._check_sizes()

    @property
    def csk_details(self) -> CSKDetails:
        """
        CSKDetails: The details object.
        """

        return self._csk_details

    @property
    def file_name(self) -> str:
        return self.csk_details.file_name


########
# base expected functionality for a module with an implemented Reader


def is_a(file_name: Union[str, BinaryIO]) -> Union[None, CSKReader]:
    """
    Tests whether a given file_name corresponds to a Cosmo Skymed file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str|BinaryIO
        the file_name to check

    Returns
    -------
    CSKReader|None
        `CSKReader` instance if Cosmo Skymed file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    if not is_hdf5(file_name):
        return None

    if h5py is None:
        return None

    try:
        csk_details = CSKDetails(file_name)
        logger.info('File {} is determined to be a Cosmo Skymed file.'.format(file_name))
        return CSKReader(csk_details)
    except SarpyIOError:
        return None



"""
Tools for inspecting a SICD urn url and providing basic details.
"""

__classification__ = 'UNCLASSIFIED'
__author__ = "Thomas McCullough"

import os
import re
from typing import List, Tuple, Dict, Optional


_SICD_DEFAULT_TUPLE = (1, 3, 0)
_SICD_SPECIFICATION_IDENTIFIER = 'SICD Volume 1 Design & Implementation Description Document'

_the_directory = os.path.split(__file__)[0]
urn_mapping = {
    'urn:SICD:0.3.1': {
        'tuple': (0, 3, 1),
        'version': '0.3.1',
        'release': '0.3.1',
        'date': '2009-03-17T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V0.3.1_2009_03_17.xsd')},
    'urn:SICD:0.4.0': {
        'tuple': (0, 4, 0),
        'version': '0.4.0',
        'release': '0.4.0',
        'date': '2010-02-12T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V0.4.0_2010_02_12.xsd')},
    'urn:SICD:0.4.1': {
        'tuple': (0, 4, 1),
        'version': '0.4.1',
        'release': '0.4.1',
        'date': '2010-07-15T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V0.4.1_2010_07_15.xsd')},
    'urn:SICD:0.5.0': {
        'tuple': (0, 5, 0),
        'version': '0.5.0',
        'release': '0.5.0',
        'date': '2011-01-12T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V0.5.0_2011_01_12.xsd')},
    'urn:SICD:1.0.0': {
        'tuple': (1, 0, 0),
        'version': '1.0.0',
        'release': '1.0.0',
        'date': '2011-08-31T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V1.0.0_2011_08_31.xsd')},
    'urn:SICD:1.0.1': {
        'tuple': (1, 0, 1),
        'version': '1.0.1',
        'release': '1.0.1',
        'date': '2013-02-25T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V1.0.1_2013_02_25.xsd')},
    'urn:SICD:1.1.0': {
        'tuple': (1, 1, 0),
        'version': '1.1.0',
        'release': '1.1.0',
        'date': '2014-09-30T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V1.1.0_2014_09_30.xsd')},
    'urn:SICD:1.2.0': {
        'tuple': (1, 2, 0),
        'version': '1.2.0',
        'release': '1.2.0',
        'date': '2016-06-30T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V1.2.0_2016_06_30.xsd')},
    'urn:SICD:1.2.1': {
        'tuple': (1, 2, 1),
        'version': '1.2.1',
        'release': '1.2.1',
        'date': '2018-12-13T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V1.2.1_2018_12_13.xsd')},
    'urn:SICD:1.3.0': {
        'tuple': (1, 3, 0),
        'version': '1.3.0',
        'release': '1.3.0',
        'date': '2021-11-30T00:00:00Z',
        'schema': os.path.join(_the_directory, 'SICD_schema_V1.3.0_2021_11_30.xsd')}
}
WRITABLE_VERSIONS = tuple(entry['release'] for key, entry in urn_mapping.items() if entry['tuple'] >= (1, 0, 0))

# validate the defined paths
for key, entry in urn_mapping.items():
    schema_path = entry.get('schema', None)
    if schema_path is not None and not os.path.exists(schema_path):
        raise ValueError('`{}` has nonexistent schema path {}'.format(key, schema_path))


def get_default_tuple() -> Tuple[int, int, int]:
    """
    Get the default SICD version tuple.

    Returns
    -------
    Tuple[int, int, int]
    """

    return _SICD_DEFAULT_TUPLE


def get_default_version_string() -> str:
    """
    Get the default SICD version string.

    Returns
    -------
    str
    """

    return '{}.{}.{}'.format(*_SICD_DEFAULT_TUPLE)


def get_specification_identifier() -> str:
    """
    Get the SICD specification identifier string.

    Returns
    -------
    str
    """

    return _SICD_SPECIFICATION_IDENTIFIER


def check_urn(urn_string: str) -> str:
    """
    Checks that the urn string follows the correct pattern.

    Parameters
    ----------
    urn_string : str

    Returns
    -------
    str

    Raises
    ------
    ValueError
        This raises an exception for a poorly formed or unmapped SICD urn.
    """

    if not isinstance(urn_string, str):
        raise TypeError(
            'Expected a urn input of string type, got type {}'.format(type(urn_string)))

    the_match = re.match(r'^\d.\d.\d$', urn_string)
    if the_match is not None:
        urn_string = 'urn:SICD:{}'.format(urn_string)

    the_match = re.match(r'^urn:SICD:\d.\d.\d$', urn_string)
    if the_match is None:
        raise ValueError(
            'Input provided as `{}`,\nbut should be of the form '
            '`urn:SICD:<major>.<minor>.<release>'.format(urn_string))
    return urn_string


def get_urn_details(urn_string: str) -> Dict[str, str]:
    """
    Gets the associated details for the given SICD urn, or raise an exception for
    poorly formatted or unrecognized urn.

    Parameters
    ----------
    urn_string : str

    Returns
    -------
    Dict[str, str]
    """

    urn_string = check_urn(urn_string)
    out = urn_mapping.get(urn_string, None)

    if out is None:
        raise KeyError(
            'Got correctly formatted, but unmapped SICD urn {}.'.format(urn_string))
    return out


def get_schema_path(the_urn: str) -> Optional[str]:
    """
    Gets the path to the proper schema file for the given urn.

    Parameters
    ----------
    the_urn : str

    Returns
    -------
    None|str
    """

    result = get_urn_details(the_urn)
    return result.get('schema', None)


def get_versions() -> List[str]:
    """
    Gets a list of recognized SICD urn.

    Returns
    -------
    List[str]
    """

    return list(sorted(urn_mapping.keys(), key=lambda x: urn_mapping[x]['tuple']))



"""
The AnnotationsType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
from typing import Union, List

import numpy

from .base import DEFAULT_STRICT, FLOAT_FORMAT

from sarpy.io.xml.base import Serializable, create_new_node, create_text_node, \
    get_node_value, find_first_child, find_children
from sarpy.io.xml.descriptors import SerializableDescriptor, SerializableListDescriptor, \
    FloatDescriptor, StringDescriptor, StringListDescriptor
from sarpy.geometry.geometry_elements import Point as PointType, \
    LineString as LineStringType, \
    LinearRing as LinearRingType, \
    Polygon as PolygonType, \
    MultiPoint as MultiPointType, \
    MultiLineString as MultiLineStringType, \
    MultiPolygon as MultiPolygonType

logger = logging.getLogger(__name__)


class ParameterType(Serializable):
    """
    The parameter type.
    """

    _fields = ('ParameterName', 'Value')
    _required = _fields
    _numeric_format = {'Value': FLOAT_FORMAT}
    # Descriptor
    ParameterName = StringDescriptor(
        'ParameterName', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    Value = FloatDescriptor(
        'Value', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: float

    def __init__(self, ParameterName=None, Value=None, **kwargs):
        """

        Parameters
        ----------
        ParameterName : str
        Value : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ParameterName = ParameterName
        self.Value = Value
        super(ParameterType, self).__init__(**kwargs)


class ProjectionType(Serializable):
    """
    The projection type.
    """

    _fields = ('ProjectionName', )
    _required = ('ProjectionName', )
    # Descriptor
    ProjectionName = StringDescriptor(
        'ProjectionName', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str

    def __init__(self, ProjectionName=None, **kwargs):
        """

        Parameters
        ----------
        ProjectionName : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ProjectionName = ProjectionName
        super(ProjectionType, self).__init__(**kwargs)


class PrimeMeridianType(Serializable):
    """
    The prime meridian location.
    """

    _fields = ('Name', 'Longitude')
    _required = _fields
    _numeric_format = {'Longitude': FLOAT_FORMAT}
    # Descriptor
    Name = StringDescriptor(
        'Name', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    Longitude = FloatDescriptor(
        'Longitude', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: float

    def __init__(self, Name=None, Longitude=None, **kwargs):
        """

        Parameters
        ----------
        Name : str
        Longitude : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Name = Name
        self.Longitude = Longitude
        super(PrimeMeridianType, self).__init__(**kwargs)


class SpheroidType(Serializable):
    """

    """
    _fields = ('SpheroidName', 'SemiMajorAxis', 'InverseFlattening')
    _required = _fields
    _numeric_format = {'SemiMajorAxis': FLOAT_FORMAT, 'InverseFlattening': FLOAT_FORMAT}
    # Descriptor
    SpheroidName = StringDescriptor(
        'SpheroidName', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    SemiMajorAxis = FloatDescriptor(
        'SemiMajorAxis', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: float
    InverseFlattening = FloatDescriptor(
        'InverseFlattening', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: float

    def __init__(self, SpheroidName=None, SemiMajorAxis=None, InverseFlattening=None, **kwargs):
        """

        Parameters
        ----------
        SpheroidName : str
        SemiMajorAxis : float
        InverseFlattening : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SpheroidName = SpheroidName
        self.SemiMajorAxis = SemiMajorAxis
        self.InverseFlattening = InverseFlattening
        super(SpheroidType, self).__init__(**kwargs)


class DatumType(Serializable):
    """

    """
    _fields = ('Spheroid', )
    _required = ('Spheroid', )
    # Descriptor
    Spheroid = SerializableDescriptor(
        'Spheroid', SpheroidType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: SpheroidType

    def __init__(self, Spheroid=None, **kwargs):
        """

        Parameters
        ----------
        Spheroid : SpheroidType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Spheroid = Spheroid
        super(DatumType, self).__init__(**kwargs)


class GeographicCoordinateSystemType(Serializable):
    """

    """
    _fields = ('Csname', 'Datum', 'PrimeMeridian', 'AngularUnit', 'LinearUnit')
    _required = ('Csname', 'Datum', 'PrimeMeridian', 'AngularUnit')
    # Descriptor
    Csname = StringDescriptor(
        'Csname', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    Datum = SerializableDescriptor(
        'Datum', DatumType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: DatumType
    PrimeMeridian = SerializableDescriptor(
        'PrimeMeridian', PrimeMeridianType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PrimeMeridianType
    AngularUnit = StringDescriptor(
        'AngularUnit', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    LinearUnit = StringDescriptor(
        'LinearUnit', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str

    def __init__(self, Csname=None, Datum=None, PrimeMeridian=None, AngularUnit=None, LinearUnit=None, **kwargs):
        """

        Parameters
        ----------
        Csname : str
        Datum : DatumType
        PrimeMeridian : PrimeMeridianType
        AngularUnit : str
        LinearUnit : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Csname = Csname
        self.Datum = Datum
        self.PrimeMeridian = PrimeMeridian
        self.AngularUnit = AngularUnit
        self.LinearUnit = LinearUnit
        super(GeographicCoordinateSystemType, self).__init__(**kwargs)


class ProjectedCoordinateSystemType(Serializable):
    """

    """
    _fields = ('Csname', 'GeographicCoordinateSystem', 'Projection', 'Parameter', 'LinearUnit')
    _required = ('Csname', 'GeographicCoordinateSystem', 'Projection', 'LinearUnit')
    # Descriptor
    Csname = StringDescriptor(
        'Csname', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    GeographicCoordinateSystem = SerializableDescriptor(
        'GeographicCoordinateSystem', GeographicCoordinateSystemType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: GeographicCoordinateSystemType
    Projection = SerializableDescriptor(
        'Projection', ProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: ProjectionType
    Parameter = SerializableDescriptor(
        'Parameter', ParameterType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, ParameterType]
    LinearUnit = StringDescriptor(
        'LinearUnit', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str

    def __init__(self, Csname=None, GeographicCoordinateSystem=None, Projection=None, Parameter=None,
                 LinearUnit=None, **kwargs):
        """

        Parameters
        ----------
        Csname : str
        GeographicCoordinateSystem : GeographicCoordinateSystemType
        Projection : ProjectionType
        Parameter : None|ParameterType
        LinearUnit : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Csname = Csname
        self.GeographicCoordinateSystem = GeographicCoordinateSystem
        self.Projection = Projection
        self.Parameter = Parameter
        self.LinearUnit = LinearUnit
        super(ProjectedCoordinateSystemType, self).__init__(**kwargs)


class GeocentricCoordinateSystemType(Serializable):
    """

    """
    _fields = ('Name', 'Datum', 'PrimeMeridian', 'LinearUnit')
    _required = ('Name', 'Datum', 'PrimeMeridian', 'LinearUnit')
    # Descriptor
    Name = StringDescriptor(
        'Name', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    Datum = SerializableDescriptor(
        'Datum', DatumType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: DatumType
    PrimeMeridian = SerializableDescriptor(
        'PrimeMeridian', PrimeMeridianType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PrimeMeridianType
    LinearUnit = StringDescriptor(
        'LinearUnit', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str

    def __init__(self, Name=None, Datum=None, PrimeMeridian=None, LinearUnit=None, **kwargs):
        """

        Parameters
        ----------
        Name : str
        Datum : DatumType
        PrimeMeridian : PrimeMeridianType
        LinearUnit : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Name = Name
        self.Datum = Datum
        self.PrimeMeridian = PrimeMeridian
        self.LinearUnit = LinearUnit
        super(GeocentricCoordinateSystemType, self).__init__(**kwargs)


class ReferenceSystemType(Serializable):
    """
    The reference system.
    """
    _fields = ('ProjectedCoordinateSystem', 'GeographicCoordinateSystem', 'GeocentricCoordinateSystem', 'AxisNames')
    _required = ('ProjectedCoordinateSystem', 'GeographicCoordinateSystem', 'GeocentricCoordinateSystem', 'AxisNames')
    _collections_tags = {'AxisNames': {'array': False, 'child_tag': 'AxisName'}}
    # Descriptor
    ProjectedCoordinateSystem = SerializableDescriptor(
        'ProjectedCoordinateSystem', ProjectedCoordinateSystemType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: ProjectedCoordinateSystemType
    GeographicCoordinateSystem = SerializableDescriptor(
        'GeographicCoordinateSystem', GeographicCoordinateSystemType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: GeographicCoordinateSystemType
    GeocentricCoordinateSystem = SerializableDescriptor(
        'GeocentricCoordinateSystem', GeocentricCoordinateSystemType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: GeocentricCoordinateSystemType
    AxisNames = StringListDescriptor(
        'AxisNames', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[str]

    def __init__(self, ProjectedCoordinateSystem=None, GeographicCoordinateSystem=None,
                 GeocentricCoordinateSystem=None, AxisNames=None, **kwargs):
        """

        Parameters
        ----------
        ProjectedCoordinateSystem : ProjectedCoordinateSystemType
        GeographicCoordinateSystem : GeographicCoordinateSystemType
        GeocentricCoordinateSystem : GeocentricCoordinateSystemType
        AxisNames : List[str]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ProjectedCoordinateSystem = ProjectedCoordinateSystem
        self.GeographicCoordinateSystem = GeographicCoordinateSystem
        self.GeocentricCoordinateSystem = GeocentricCoordinateSystem
        self.AxisNames = AxisNames
        super(ReferenceSystemType, self).__init__(**kwargs)


class AnnotationObjectType(Serializable):
    """
    Geometrical representation of the annotation. Only one of the geometry elements should be populated.
    It will not be enforced, but the order of preference will be `('Point', 'Line', 'LinearRing', 'Polygon',
    'MultiPoint', 'MultiLineString', 'MultiPolygon')`.

    Note that PolyhedralSurface is not currently supported.
    """
    _fields = ('Point', 'Line', 'LinearRing', 'Polygon', 'MultiPoint', 'MultiLineString', 'MultiPolygon')
    _choice = ({'required': True, 'collection': _fields}, )

    def __init__(self, Point=None, Line=None, LinearRing=None, Polygon=None, MultiPoint=None,
                 MultiLineString=None, MultiPolygon=None, **kwargs):
        """

        Parameters
        ----------
        Point : sarpy.geometry.geometry_elements.Point|numpy.array|list|tuple
        Line : sarpy.geometry.geometry_elements.LineString|numpy.array|list|tuple
        LinearRing : sarpy.geometry.geometry_elements.LinearRing|numpy.array|list|tuple
        Polygon : sarpy.geometry.geometry_elements.Polygon|list
        MultiPoint : sarpy.geometry.geometry_elements.MultiPoint|list
        MultiLineString : sarpy.geometry.geometry_elements.MultiLineString|list
        MultiPolygon : sarpy.geometry.geometry_elements.MultiPolygon|list
        kwargs
        """

        self._Point = None
        self._Line = None
        self._LinearRing = None
        self._Polygon = None
        self._MultiPoint = None
        self._MultiLineString = None
        self._MultiPolygon = None

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']

        if Point is not None:
            self.Point = Point
        elif Line is not None:
            self.Line = Line
        elif LinearRing is not None:
            self.LinearRing = LinearRing
        elif Polygon is not None:
            self.Polygon = Polygon
        elif MultiPoint is not None:
            self.MultiPoint = MultiPoint
        elif MultiLineString is not None:
            self.MultiLineString = MultiLineString
        elif MultiPolygon is not None:
            self.MultiPolygon = MultiPolygon
        else:
            logger.error(
                "One of (Point, Line, LinearRing, Polygon, MultiPoint, MultiLineString, MultiPolygon)\n\t"
                "should have been provided to the AnnotationObjectType constructor.")
        super(AnnotationObjectType, self).__init__(**kwargs)

    @property
    def GeometryType(self):
        """
        str: **READ ONLY** The type of geometric element which is set, from
        ('Point', 'Line', 'LinearRing', 'Polygon', 'MultiPoint',
        'MultiLineString', 'MultiPolygon')
        """

        for attribute in self._choice[0]['collection']:
            if getattr(self, attribute) is not None:
                return attribute
        return None

    @property
    def Point(self):
        """
        None|sarpy.geometry.geometry_elements.Point: The point.
        """

        return self._Point

    @Point.setter
    def Point(self, value):
        if value is None:
            self._Point = None
        elif isinstance(value, (numpy.ndarray, list, tuple)):
            self._Point = PointType(coordinates=value)
        elif isinstance(value, PointType):
            self._Point = value
        else:
            raise TypeError(
                'Point requires and instance of sarpy.geometry.geometry_elements.Point, '
                'got {}'.format(type(value)))

    @property
    def Line(self):
        """
        None|sarpy.geometry.geometry_elements.LineString: The line.
        """

        return self._Line

    @Line.setter
    def Line(self, value):
        if value is None:
            self._Line = None
        elif isinstance(value, (numpy.ndarray, list, tuple)):
            self._Line = LineStringType(coordinates=value)
        elif isinstance(value, LineStringType):
            self._Line = value
        else:
            raise TypeError(
                'Line requires and instance of sarpy.geometry.geometry_elements.LineString, '
                'got {}'.format(type(value)))

    @property
    def LinearRing(self):
        """
        None|sarpy.geometry.geometry_elements.LinearRing: The linear ring.
        """

        return self._LinearRing

    @LinearRing.setter
    def LinearRing(self, value):
        if value is None:
            self._LinearRing = None
        elif isinstance(value, (numpy.ndarray, list, tuple)):
            self._LinearRing = LinearRingType(coordinates=value)
        elif isinstance(value, LinearRingType):
            self._LinearRing = value
        else:
            raise TypeError(
                'LinearRing requires and instance of sarpy.geometry.geometry_elements.LinearRing, '
                'got {}'.format(type(value)))

    @property
    def Polygon(self):
        """
        None|sarpy.geometry.geometry_elements.Polygon: The polygon.
        """

        return self._Polygon

    @Polygon.setter
    def Polygon(self, value):
        if value is None:
            self._Polygon = None
        elif isinstance(value, list):
            self._Polygon = PolygonType(coordinates=value)
        elif isinstance(value, PolygonType):
            self._Polygon = value
        else:
            raise TypeError(
                'Polygon requires and instance of sarpy.geometry.geometry_elements.Polygon, '
                'got {}'.format(type(value)))

    @property
    def MultiPoint(self):
        """
        None|sarpy.geometry.geometry_elements.MultiPoint: The multipoint.
        """

        return self._MultiPoint

    @MultiPoint.setter
    def MultiPoint(self, value):
        if value is None:
            self._MultiPoint = None
        elif isinstance(value, list):
            self._MultiPoint = MultiPointType(coordinates=value)
        elif isinstance(value, MultiPointType):
            self._MultiPoint = value
        else:
            raise TypeError(
                'MultiPoint requires and instance of sarpy.geometry.geometry_elements.MultiPoint, '
                'got {}'.format(type(value)))

    @property
    def MultiLineString(self):
        """
        None|sarpy.geometry.geometry_elements.MultiLineString: The multi-linestring.
        """

        return self._MultiLineString

    @MultiLineString.setter
    def MultiLineString(self, value):
        if value is None:
            self._MultiLineString = None
        elif isinstance(value, list):
            self._MultiLineString = MultiLineStringType(coordinates=value)
        elif isinstance(value, MultiLineStringType):
            self._MultiLineString = value
        else:
            raise TypeError(
                'MultiLineString requires and instance of sarpy.geometry.geometry_elements.MultiLineString, '
                'got {}'.format(type(value)))

    @property
    def MultiPolygon(self):
        """
        None|sarpy.geometry.geometry_elements.MultiPolygon: The multi-polygon.
        """

        return self._MultiPolygon

    @MultiPolygon.setter
    def MultiPolygon(self, value):
        if value is None:
            self._MultiPolygon = None
        elif isinstance(value, list):
            self._MultiPolygon = MultiPolygonType(coordinates=value)
        elif isinstance(value, MultiPolygonType):
            self._MultiPolygon = value
        else:
            raise TypeError(
                'MultiPolygon requires and instance of sarpy.geometry.geometry_elements.MultiPolygon, '
                'got {}'.format(type(value)))

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        for attr in self._fields:
            val = getattr(self, attr)
            if val is not None:
                return val.to_dict()
        return {}

    @classmethod
    def from_dict(cls, input_dict):
        typ = input_dict.get('type', None)
        if typ is None:
            return cls()
        elif typ == 'Point':
            return cls(Point=PointType.from_dict(input_dict))
        elif typ == 'LineString':
            return cls(Line=LineStringType.from_dict(input_dict))
        elif typ == 'LinearRing':
            return cls(LinearRing=LinearRingType.from_dict(input_dict))
        elif typ == 'Polygon':
            return cls(Polygon=PolygonType.from_dict(input_dict))
        elif typ == 'MultiPoint':
            return cls(MultiPoint=MultiPointType.from_dict(input_dict))
        elif typ == 'MultiLineString':
            return cls(MultiLineString=MultiLineStringType.from_dict(input_dict))
        elif typ == 'MultiPolygon':
            return cls(MultiPolygon=MultiPolygonType.from_dict(input_dict))
        else:
            logger.error(
                'AnnotationObjectType got unsupported input dictionary {}.\n\t'
                'Returning None.'.format(input_dict))
            return None

    @staticmethod
    def _serialize_point(coords, doc, tag, parent):
        if len(coords) < 2:
            raise ValueError('coords must have at least two elements')
        fmt_func = '{0:0.17E}'.format
        node = create_new_node(doc, tag, parent=parent)
        create_text_node(doc, 'sfa:X', fmt_func(coords[0]), parent=node)
        create_text_node(doc, 'sfa:Y', fmt_func(coords[1]), parent=node)
        if len(coords) > 2:
            create_text_node(doc, 'sfa:Z', fmt_func(coords[2]), parent=node)
        if len(coords) > 3:
            create_text_node(doc, 'sfa:M', fmt_func(coords[3]), parent=node)

    def _serialize_line(self, coords, doc, tag, parent):
        node = create_new_node(doc, tag, parent=parent)
        for entry in coords:
            self._serialize_point(entry, doc, 'sfa:Vertex', node)

    def _serialize_polygon(self, coords, doc, tag, parent):
        node = create_new_node(doc, tag, parent=parent)
        for entry in coords:
            self._serialize_line(entry, doc, 'sfa:Ring', node)

    def _serialize_multilinestring(self, coords, doc, tag, parent):
        node = create_new_node(doc, tag, parent=parent)
        for entry in coords:
            self._serialize_line(entry, doc, 'sfa:Element', node)

    def _serialize_multipolygon(self, coords, doc, tag, parent):
        node = create_new_node(doc, tag, parent=parent)
        for entry in coords:
            self._serialize_polygon(entry, doc, 'sfa:Element', node)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        if parent is None:
            parent = doc.getroot()
        if ns_key is None:
            node = create_new_node(doc, tag, parent=parent)
        else:
            node = create_new_node(doc, '{}:{}'.format(ns_key, tag), parent=parent)
        typ = self.GeometryType
        if typ is None:
            return node

        coords = getattr(self, typ).get_coordinate_list()
        if typ == 'Point':
            self._serialize_point(coords, doc, 'sfa:Point', node)
        elif typ == 'Line':
            self._serialize_line(coords, doc, 'sfa:Line', node)
        elif typ == 'LinearRing':
            self._serialize_line(coords, doc, 'sfa:LinearRing', node)
        elif typ == 'Polygon':
            self._serialize_polygon(coords, doc, 'sfa:Polygon', node)
        elif typ == 'MultiPoint':
            self._serialize_line(coords, doc, 'sfa:MultiPoint', node)
        elif typ == 'MultiLineString':
            self._serialize_multilinestring(coords, doc, 'sfa:MultiLineString', node)
        elif typ == 'MultiPolygon':
            self._serialize_multipolygon(coords, doc, 'sfa:MultiPolygon', node)
        else:
            raise ValueError('Unsupported serialization type {}'.format(typ))
        return node

    @staticmethod
    def _get_value(node, tag, xml_ns, ns_key):
        t_node = find_first_child(node, tag, xml_ns, ns_key)
        if t_node is None:
            return None
        else:
            return float(get_node_value(t_node))

    @classmethod
    def _extract_point(cls, node, xml_ns):
        out = [cls._get_value(node, 'X', xml_ns, 'sfa'), cls._get_value(node, 'Y', xml_ns, 'sfa')]
        z = cls._get_value(node, 'Z', xml_ns, 'sfa')
        if z in None:
            return out
        out.append(z)
        m = cls._get_value(node, 'M', xml_ns, 'sfa')
        if m in None:
            return out
        out.append(m)
        return out

    @classmethod
    def _extract_line(cls, node, xml_ns, tag='Vertex'):
        v_nodes = find_children(node, tag, xml_ns, 'sfa')
        return [cls._extract_point(v_node, xml_ns) for v_node in v_nodes]

    @classmethod
    def _extract_polygon(cls, node, xml_ns, tag='Ring'):
        v_nodes = find_children(node, tag, xml_ns, 'sfa')
        return [cls._extract_line(v_node, xml_ns, tag='Vertex') for v_node in v_nodes]

    @classmethod
    def _deserialize_point(cls, node, xml_ns, tag='Point'):
        point_node = find_first_child(node, tag, xml_ns, 'sfa')
        if point_node is None:
            return None
        return cls._extract_point(point_node, xml_ns)

    @classmethod
    def _deserialize_line(cls, node, xml_ns, tag='Line'):
        line_node = find_first_child(node, tag, xml_ns, 'sfa')
        if line_node is None:
            return None
        return cls._extract_line(line_node, xml_ns, tag='Vertex')

    @classmethod
    def _deserialize_polygon(cls, node, xml_ns, tag='Polygon'):
        poly_node = find_first_child(node, tag, xml_ns, 'sfa')
        if poly_node is None:
            return None
        return cls._extract_polygon(poly_node, xml_ns, tag='Ring')

    @classmethod
    def _deserialize_multilinestring(cls, node, xml_ns, tag='MultiLineString'):
        mls_node = find_first_child(node, tag, xml_ns, 'sfa')
        if mls_node is None:
            return None
        ls_nodes = find_children(mls_node, 'Element', xml_ns, 'sfa')
        return [cls._extract_line(ls_node, xml_ns, tag='Vertex') for ls_node in ls_nodes]

    @classmethod
    def _deserialize_multipolygon(cls, node, xml_ns, tag='MultiPolygon'):
        mp_node = find_first_child(node, tag, xml_ns, 'sfa')
        if mp_node is None:
            return None
        p_nodes = find_children(mp_node, 'Element', xml_ns, 'sfa')
        return [cls._extract_polygon(p_node, xml_ns, tag='Ring') for p_node in p_nodes]

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if xml_ns is None or 'sfa' not in xml_ns:
            raise ValueError('xml_ns must contain an entry for key "sfa"')

        coords = cls._deserialize_point(node, xml_ns, 'Point')
        if coords is not None:
            return cls(Point=PointType(coordinates=coords))

        coords = cls._deserialize_line(node, xml_ns, tag='Line')
        if coords is not None:
            return cls(Line=LineStringType(coordinates=coords))

        coords = cls._deserialize_line(node, xml_ns, tag='LinearRing')
        if coords is not None:
            return cls(LinearRing=LinearRingType(coordinates=coords))

        coords = cls._deserialize_polygon(node, xml_ns, tag='Polygon')
        if coords is not None:
            return cls(Polygon=PolygonType(coordinates=coords))

        coords = cls._deserialize_line(node, xml_ns, tag='MultiPoint')
        if coords is not None:
            return cls(MultiPoint=MultiPointType(coordinates=coords))

        coords = cls._deserialize_multilinestring(node, xml_ns, tag='MultiLineString')
        if coords is not None:
            return cls(MultiLineString=MultiLineStringType(coordinates=coords))

        coords = cls._deserialize_multipolygon(node, xml_ns, tag='MultiPolygon')
        if coords is not None:
            return cls(MultiPolygon=MultiPolygonType(coordinates=coords))

        return cls()


class AnnotationType(Serializable):
    """
    The annotation type.
    """

    _fields = ('Identifier', 'SpatialReferenceSystem', 'Objects')
    _required = ('Identifier', 'Objects')
    _collections_tags = {'Objects': {'array': False, 'child_tag': 'Object'}}
    # Descriptor
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    SpatialReferenceSystem = SerializableDescriptor(
        'SpatialReferenceSystem', ReferenceSystemType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, ReferenceSystemType]
    Objects = SerializableListDescriptor(
        'Objects', AnnotationObjectType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[AnnotationObjectType]

    def __init__(self, Identifier=None, SpatialReferenceSystem=None, Objects=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        SpatialReferenceSystem : None|ReferenceSystemType
        Objects : List[AnnotationObjectType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.SpatialReferenceSystem = SpatialReferenceSystem
        self.Objects = Objects
        super(AnnotationType, self).__init__(**kwargs)


class AnnotationsType(Serializable):
    """
    The list of annotations.
    """

    _fields = ('Annotations', )
    _required = ('Annotations', )
    _collections_tags = {'Annotations': {'array': False, 'child_tag': 'Annotation'}}
    # Descriptor
    Annotations = SerializableListDescriptor(
        'Annotations', AnnotationType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[AnnotationType]

    def __init__(self, Annotations=None, **kwargs):
        """

        Parameters
        ----------
        Annotations : List[AnnotationType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Annotations = Annotations
        super(AnnotationsType, self).__init__(**kwargs)

    def __len__(self):
        return len(self.Annotations)

    def __getitem__(self, item):
        return self.Annotations[item]




__classification__ = 'UNCLASSIFIED'

from .base import FullResolutionFetcher, OrthorectificationIterator
from .ortho_methods import OrthorectificationHelper, NearestNeighborMethod, BivariateSplineMethod
from .projection_helper import ProjectionHelper, PGProjection, PGRatPolyProjection
